package se2030.studentroster;

import java.util.ArrayList;
import java.util.List;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import se2030.parser.AbstractParserEventHandler;

/*
 * This class provides a sample concrete implementation of the AbstractParserEventHandler,
 * which is an abstract class that provides default implementations of the
 * event handling methods that are called by the Parser.
 * 
 * In this implementation, this handler builds a roster (a collection of Students) based on the 
 * following XML representation:
 * <?xml version="1.0" encoding="UTF-8"?>

  <roster>
  
    <name>SE2030 Fall 2019</name>
    <!-- here is a comment -->
    
     <section>
     
      <student firstname="Bill" lastname="Bored">
        <program>SE</program>
        <!-- another comment -->
        <id>1111</id>
      </student>
      
      <student lastname="Sledd" firstname="Bob">
        <program>CE</program>
        <id>1111</id>
      </student>
      
    </section>
  </roster>


 *  
 */
public class RosterBuilder extends AbstractParserEventHandler {
	// possible state values
	private enum PossibleStates {INITIAL, ROSTER, NAME, SECTION, STUDENT, PROGRAM, ID, FINAL}; // indentation level
	private PossibleStates currentState = PossibleStates.INITIAL; // starting state of parsing
	
	// We need to declare some kind of Student collection class here.
	private List<Student> roster; // this will hold a collection of Student objects
	private String name; // name of this roster - see above XML sample
	private Student currentStudent; // the Student object currently being "assembled" from XML
	
	public RosterBuilder() {
		super();
	}
	
	public List<Student> getRoster() {
		return roster;
	}
		
	public String getName() {
		return name;
	}

	@Override
	// This method is called whenever the Parser encounters a new element tag
	// which may or may not have Attributes. For example,
	//       <student firstname="Bill" lastname="Bored">
	//   or  
	//           <name>SE2030 Roster</name>
	//
	// The <student> element contains two attributes: firstname and lastname.
	// Conversely, <name>SE2030 Roster</name> contains no attributes, but it does contain text.
	// The code below shows how to access the values of the attributes when they exist.
	// The text within an element is accessed when the characters() method below is called by the Parser.
	public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
		// localName contains the name of the element - e.g. roster, name, section, student etc
		
		int line = locator.getLineNumber(); // current line being parsed
		int column = locator.getColumnNumber(); // current column being parsed
		
		log("startElement found", "<"+localName+"> at line " + line + ", col " + column ); // debug logger

		// We need to ensure that elements are found in the correct order by maintaining state.
		// For instance:
		if( currentState == PossibleStates.INITIAL ) {
			if( !localName.equalsIgnoreCase("roster") ) { // <roster> should be the first element found
				throw new SAXException("Expected <roster> element at this location! line "+line + ", col "+column);
				// We must make sure that no other element besides <roster> is at the start of the xml file;
				// if some other element is found, we have an invalid file, so throw a SAXException.
			}
			currentState = PossibleStates.ROSTER; // once <roster> is found, we're in the ROSTER state!
			
			// Now that we've found a <roster> element, we know we have a valid file. We can init our Student collection here.
			roster = new ArrayList<>();
		} 
	
		if( localName.equalsIgnoreCase("name")) {
			if( currentState != PossibleStates.ROSTER ) { // <name> should only be found within <roster>
				throw new SAXException("<name> element found in illegal location!");
			}
			currentState = PossibleStates.NAME; // once <name> is found, we're in the NAME state!
		}
		
		if( localName.equalsIgnoreCase("section")) {
			if( currentState != PossibleStates.ROSTER ) { // <section> should only be found within <roster>
				throw new SAXException("<section> element found in illegal location!");
			}
			currentState = PossibleStates.SECTION; // once <section> is found, we're in the SECTION state!
		}
		
		//TODO: add other logic to check for <name> and advance to the NAME state

		if( localName.equalsIgnoreCase("student")) {
			if( currentState != PossibleStates.SECTION ) { // <student> should only be found within <section>
				throw new SAXException("<student> element found in illegal location!");
			}
			// We must advance the state to STUDENT here.
			currentState = PossibleStates.STUDENT; // once <student> is found, we're in the STUDENT state!
			
			
			// Every <student> element should have 2 attributes (firstname and lastname)
			if( atts.getLength() != 2 ) { 
				throw new SAXException("<student> element has an illegal number of attributes: " + atts.getLength() );
			}

			// We must check to make sure that the <student> element has both firstname and lastname
			// attributes, so that we can set those values in the Student constructor below.
			// If either of those attributes don't exist or are invalid, we throw a SAXException.
			String firstAtt = atts.getLocalName(0); // should be "firstname" or "lastname"
			String firstAttValue = atts.getValue(0); 
			String secondAtt = atts.getLocalName(1); // should be "lastname" or "firstname"
			String secondAttValue = atts.getValue(1); 

			String firstname, lastname;
			
			// attributes in normal order
			if( firstAtt.equalsIgnoreCase("firstname") &&  secondAtt.equalsIgnoreCase("lastname") ) {
				firstname = firstAttValue;
				lastname = secondAttValue;
			}	
			// attributes in reverse order
			else if( firstAtt.equalsIgnoreCase("lastname") &&  secondAtt.equalsIgnoreCase("firstname") ) {
				firstname = secondAttValue;
				lastname = firstAttValue;
			}
			// incorrect attributes!
			else
				throw new SAXException("<student> element has one or more illegal attributes: " + firstAtt + ":" + secondAtt );


			// We can now create/initialize a new Student object now that we have determined valid attributes.
			currentStudent = new Student(firstname, lastname);
			// We will add the program and id to the Student when the characters() method (below) is called by the Parser.
		}
		if( localName.equalsIgnoreCase("program")) {
			if( currentState != PossibleStates.STUDENT ) { // <program> should only be found within <student>
				throw new SAXException("<program> element found in illegal location!");
			}
			currentState = PossibleStates.PROGRAM; // once <program> is found, we're in the PROGRAM state!
		}
		if( localName.equalsIgnoreCase("id")) {
			if( currentState != PossibleStates.STUDENT ) { // <id> should only be found within <student>
				throw new SAXException("<id> element found in illegal location!");
			}
			currentState = PossibleStates.ID; // once <id> is found, we're in the ID state!
		}

	}

	@Override
	// This ContentHandler method is called whenever the Parser encounters an end tag (e.g. </trkpt>, </gpx>)
	public void endElement(String uri, String localName, String qName)
			throws SAXException {
		
		int line = locator.getLineNumber(); // current line being parsed
		int column = locator.getColumnNumber(); // current column being parsed

		
		log("endElement found", "<"+localName+"> at line " + line + ", col " + column ); // debug logger
		
		// localName contains the name of the element - e.g. roster, name, section, student etc
		if( localName.equalsIgnoreCase("id") ) {
			if( currentState != PossibleStates.ID ) { // should be back in ID state when </id> is encountered
				throw new SAXException("</id> element found in illegal location! "+line + ", col "+column );
			}
			currentState = PossibleStates.STUDENT; // </id> signals that we're back in the STUDENT state!
		}
		
		if( localName.equalsIgnoreCase("program") ) {
			if( currentState != PossibleStates.PROGRAM ) { // should be back in PROGRAM state when </PROGRAM> is encountered
				throw new SAXException("</program> element found in illegal location!");
			}
			currentState = PossibleStates.STUDENT; // </program> signals that we're back in the STUDENT state!
		}
		
		
		if( localName.equalsIgnoreCase("student") ) {
			if( currentState != PossibleStates.STUDENT ) { // should be back in STUDENT state when </student> is encountered
				throw new SAXException("</student> element found in illegal location!");
			}
			// We must make sure that we've completed the Student object - that is, the <program> and <id>
			// elements must have been encountered and their values must have been added to the Student object.
			if( currentStudent.getId() == -1 ) // -1 means that the id was never set!				
				throw new SAXException("<student> element is missing an <id> subelement or attribute! line "+line + ", col "+column );
			if( currentStudent.getProgram().isEmpty() ) // -1 means that the id was never set!				
				throw new SAXException("<student> element is missing an <program> subelement or attribute! line "+line + ", col "+column );

			// if everything is OK, add the Student to the list!
			roster.add(currentStudent);
			
			currentState = PossibleStates.SECTION; // </student> signals that we're back in the SECTION state!
		}
		
		if( localName.equalsIgnoreCase("section") ) {
			if( currentState != PossibleStates.SECTION ) { // should be back in SECTION state when </section> is encountered
				throw new SAXException("</section> element found in illegal location!");
			}
			currentState = PossibleStates.ROSTER; // </section> signals that we're back in the ROSTER state!
		}
		
		if( localName.equalsIgnoreCase("name") ) {
			if( currentState != PossibleStates.NAME ) { // should be back in NAME state when </name> is encountered
				throw new SAXException("</name> element found in illegal location!");
			}
			currentState = PossibleStates.ROSTER; // </section> signals that we're back in the ROSTER state!
		}

		if( localName.equalsIgnoreCase("roster") ) {
			if( currentState != PossibleStates.ROSTER ) { // should be back in ROSTER state when </roster> is encountered
				throw new SAXException("</roster> element found in illegal location!");
			}
			currentState = PossibleStates.FINAL; // </roster> signals that we've reached the end state!
			// When the FINAL state is reached, our collection of Students should be complete.
			// The collection can be retrieved with the getRoster() method (above).
		}
	}

	@Override
	// This ContentHandler method is called whenever the Parser encounters text between tags.
	// For example, for the tag  <name>SE2030 Fall 2019</name>, the characters "SE2030 Fall 2019" are reported.
	// For tags like <student> and <roster>, a newline character is reported.
	public void characters(char[] ch, int start, int length) throws SAXException {
		// convert the char[] array to a String for convenience
		String s = new String(ch);
		s = s.substring(start, start+length); 
		s = s.trim(); // remove leading and trailing whitespace

		if( !s.isEmpty() ) { // ignore whitespace
			int line = locator.getLineNumber(); // current line being parsed
			int column = locator.getColumnNumber(); // current column being parsed
			log("characters found", "'"+s + "' at line " + line + ", col " + column );
		}
		
		if( currentState == PossibleStates.NAME ) { // We're in the NAME state, so these are the chars found between <name> and </name>
			name = s;
		}
		
		if( currentState == PossibleStates.ID ) {// We're in the ID state, so these are the chars found between <id> and </id>
			try {
				int id = Integer.parseInt(s); // try parsing the characters as an int
				currentStudent.setId(id); // set the current Student object's id attribute
			} catch( NumberFormatException e ) {
				throw new SAXException("</id> attribute is not an integer!");
			}
		}
		if( currentState == PossibleStates.PROGRAM ) {// We're in the PROGRAM state, so these are the chars found between <program> and </program>
			if( s.isEmpty() ) 
				throw new SAXException("</program> attribute is empty!");
			currentStudent.setProgram(s);
		}
	}
	
	@Override
	//This ContentHandler method is called when the Parser reaches the end of the
	// XML document. If the document was a well-formed Student file, then the currentState of the
	// system should be possibleStates.FINAL (since the end </roster> element was
	// found). If the currentState is not FINAL here, then something was wrong
	// with the document structure, so throw a SAXException!
	public void endDocument() throws SAXException {
		log("end of document found", "parsing complete" ); // debug logger

		if( currentState != PossibleStates.FINAL )
			throw new SAXException("Document structure error. Not in FINAL state at the end of the document!");
	}

}